def main():
    # TODO: Put entry-point code here!
    print("No user code have been added to that project yet")
    print(
"To start some of built-in liteServers:\n"\
"python -m liteserver.device.litePeakSimulator\n"\
"python -m liteserver.device.liteScaler\n"\
"python -m liteserver.device.liteLabjack\n"\
"etc..."\
)
if __name__ == "__main__":
    main()
